//-------//-------//-------//-------//-------//-------//-------//-------
//
//
// edibfill.cpp : Ƃ˂ : http://www.youtube.com/watch?v=AMaNxyHhH6A
//
//
// cl edibfill.cpp /fp:fast /Ox /arch:SSE2
// cl edibfill.cpp /openmp /fp:fast /arch:SSE2 /Ox
//
//-------//-------//-------//-------//-------//-------//-------//-------
//#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <windows.h>

#pragma comment(lib, "kernel32.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "winmm.lib")

//-------//-------//-------//-------//-------//-------//-------//-------
// window  proc
//-------//-------//-------//-------//-------//-------//-------//-------
static LRESULT WINAPI SimpleMsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
  switch( msg ) {
    //need 
    case WM_SYSCOMMAND:
      switch( ( wParam & 0xFFF0 ) ) {
        case SC_MONITORPOWER:
        case SC_SCREENSAVE:
          return 0;
          break;
      }
      break;
    case WM_CLOSE:
    case WM_DESTROY:
      PostQuitMessage( 0 );
      return 0;
      break;
    case WM_IME_SETCONTEXT:
      lParam &= ~ISC_SHOWUIALL;
      break;
    case WM_KEYDOWN:
      if(wParam == VK_ESCAPE) PostQuitMessage(0);
      break;
  }
  return DefWindowProc( hWnd, msg, wParam, lParam );
}

//-------//-------//-------//-------//-------//-------//-------//-------
//
//  main loop
//
//-------//-------//-------//-------//-------//-------//-------//-------
HRESULT start(const char *szAppName, int BmpX, int BmpY, float screenratio, void (*update)(unsigned long *dest, int width, int height)) {
  HWND         hWnd        = NULL;
  HINSTANCE    hInst       = GetModuleHandle(NULL);;
  HRESULT      hRet        = E_FAIL;
  DWORD        style       = WS_OVERLAPPEDWINDOW & ~WS_MAXIMIZEBOX & ~WS_THICKFRAME;
  DWORD        dwExStyle   = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
  DWORD        Width       = GetSystemMetrics( SM_CXSCREEN ) * screenratio;
  DWORD        Height      = GetSystemMetrics( SM_CYSCREEN ) * screenratio;
  POINT        point       = {(GetSystemMetrics( SM_CXSCREEN ) - Width ) / 2, (GetSystemMetrics( SM_CYSCREEN ) - Height) / 2};
  RECT         rc          = {0, 0, Width, Height};
  BITMAPINFO   bi;
  HBITMAP      hDIB        = NULL;
  HDC          hDIBDC      = NULL;
  void       *pBits        = NULL;
  MSG          msg;
  BOOL         bActive     = TRUE;

  //RegisterClassEx
  WNDCLASSEX twc = {
    sizeof( WNDCLASSEX ), CS_CLASSDC | CS_VREDRAW | CS_HREDRAW, SimpleMsgProc, 0L, 0L,
    hInst,  LoadIcon(NULL, IDI_APPLICATION), LoadCursor( NULL , IDC_ARROW ), 
    ( HBRUSH )GetStockObject( BLACK_BRUSH ), NULL, szAppName, NULL
  };
  if( !RegisterClassEx( &twc ) ) return hRet;
  
  //CreateWindowEx
  hWnd = CreateWindowEx(dwExStyle, szAppName, szAppName, style, point.x, point.y, rc.right - rc.left, rc.bottom - rc.top, NULL, NULL, hInst, NULL );
  if(!hWnd) return hRet;
  ShowWindow(hWnd, SW_SHOW);
  UpdateWindow(hWnd);
  
  //Create Bitmap
  bi.bmiHeader.biSize        = sizeof(BITMAPINFOHEADER);
  bi.bmiHeader.biWidth       = BmpX;
  bi.bmiHeader.biHeight      = -BmpY;
  bi.bmiHeader.biPlanes      = 1;
  bi.bmiHeader.biBitCount    = 32;
  bi.bmiHeader.biCompression = BI_RGB;

  AdjustWindowRectEx( &rc, style, FALSE, dwExStyle);
  if( !(hDIB = CreateDIBSection(NULL, &bi, DIB_RGB_COLORS, (void **)&pBits, 0, 0)) ) return E_FAIL;
  hDIBDC = CreateCompatibleDC(NULL);
  SaveDC(hDIBDC);
  SelectObject(hDIBDC, hDIB);

  //start loop
  timeBeginPeriod(1);
  while(bActive) {
    while(PeekMessage(&msg , 0 , 0 , 0 , PM_REMOVE)) {
      if(msg.message == WM_QUIT) {
        bActive = FALSE;
        break;
      } else {
        TranslateMessage( &msg );
        DispatchMessage( &msg );
      }
    }
    if(GetAsyncKeyState(VK_ESCAPE) & 0x8000) bActive = FALSE;
    update((DWORD *)pBits, BmpX, BmpY);
    HDC hdc = GetDC(hWnd);
    StretchBlt(hdc, 0, 0, Width, Height, hDIBDC, 0, 0, BmpX, BmpY, SRCCOPY);
    ReleaseDC(hWnd, hdc);
    //Sleep(16);
    //Sleep(0);
  }
  timeEndPeriod(1);
  if(hDIB) DeleteObject(hDIB);
  return 0;
}


//-------//-------//-------//-------//-------//-------//-------//-------
// vec3
//-------//-------//-------//-------//-------//-------//-------//-------
struct vec3 {
  union
  {
    float v[3];
    struct { float x, y, z; };
  };
  vec3()                              { x = y = z = 0.0f; }
  vec3(float a)                       { x = y = z = a; }
  vec3(float a, float b, float c)     { x = a;    y = b;    z = c;  }
  vec3  operator +  (const vec3 &a)   { return vec3(x + a.x, y + a.y, z + a.z); }
  vec3  operator -  (const vec3 &a)   { return vec3(x - a.x, y - a.y, z - a.z); }
  vec3  operator *  (const vec3 &a)   { return vec3(x * a.x, y * a.y, z * a.z); }
  vec3  operator /  (const vec3 &a)   { return vec3(x / a.x, y / a.y, z / a.z); }
  vec3  operator +  (float a)         { return vec3(x + a, y + a, z + a); }
  vec3  operator -  (float a)         { return vec3(x - a, y - a, z - a); }
  vec3  operator *  (float a)         { return vec3(x * a, y * a, z * a); }
  vec3  operator /  (float a)         { return vec3(x / a, y / a, z / a); }
  vec3 &operator += (const vec3 &a)   { x += a.x; y += a.y; z += a.z; return *this; }
  vec3 &operator -= (const vec3 &a)   { x -= a.x; y -= a.y; z -= a.z; return *this; }
  vec3 &operator *= (const vec3 &a)   { x *= a.x; y *= a.y; z *= a.z; return *this; }
  vec3 &operator /= (const vec3 &a)   { x /= a.x; y /= a.y; z /= a.z; return *this; }
  vec3  operator >  (const vec3 &a)   { return vec3(x > a.x, y > a.y, z > a.z); }
  vec3  operator <  (const vec3 &a)   { return vec3(x < a.x, y < a.y, z < a.z); }
  float dot(vec3 &a)                  { return x * a.x + y * a.y + z * a.z; }
  vec3  cross(vec3 &a)                { return vec3( y * a.z - z * a.y, z * a.x - x * a.z, x * a.y - y * a.x); }
  float length()                      { return sqrt(dot(*this));}
  vec3 &normalize()                   { float l = 1.0f / length(); return *this = vec3(x * l, y * l, z * l); }
  float frac(float a)                 { a -= (float)(int)a; if(a < 0) return -a; return a; }
  vec3  frac()                        { return vec3(frac(x), frac(y), frac(z)); }
  vec3  abs()                         { return vec3(::abs(x), ::abs(y), ::abs(z)); }
  vec3  integer()                     { return vec3((float)(int)(x), (float)(int)(y), (float)(int)(z)); }
  vec3 &left()                        { float t = x; x = y; y = z; z = t; return *this;}
  vec3 &right()                       { float t = z; z = y; y = x; x = t; return *this;}
  DWORD color()
  {
    float r = x > 1.0f ? 1.0f : x < 0.0f ? 0.0f : x;
    float g = y > 1.0f ? 1.0f : y < 0.0f ? 0.0f : y;
    float b = z > 1.0f ? 1.0f : z < 0.0f ? 0.0f : z;
    return (int(r * 255) << 16) | (int(g * 255) << 8) | int(b * 255);
  }
  void  disp()                        { printf("%f %f %f \n", x, y, z); }
};



//-------//-------//-------//-------//-------//-------//-------//-------
// color
//-------//-------//-------//-------//-------//-------//-------//-------
union color {
  unsigned long argb;
  struct { unsigned char b, g, r, a; };
  color()                { argb = 0; }
  color(unsigned long c) { argb = c; }
  color &blend(color &c, int k) {
    int k1 = k;
    int k2 = 256 - k1;
    a = ((int(a) * k1) + (int(c.a) * k2)) >> 8;
    r = ((int(r) * k1) + (int(c.r) * k2)) >> 8;
    g = ((int(g) * k1) + (int(c.g) * k2)) >> 8;
    b = ((int(b) * k1) + (int(c.b) * k2)) >> 8;
    return *this;
  }
  
  color &blend(color &c, int k, int l, int m, int n) {
    int k1 = k; int k2 = 256 - k1;
    int l1 = l; int l2 = 256 - l1;
    int m1 = m; int m2 = 256 - m1;
    int n1 = n; int n2 = 256 - n1;
    a = ((int(a) * k1) + (int(c.a) * k2)) >> 8;
    r = ((int(r) * l1) + (int(c.r) * l2)) >> 8;
    g = ((int(g) * m1) + (int(c.g) * m2)) >> 8;
    b = ((int(b) * n1) + (int(c.b) * n2)) >> 8;
    return *this;
  }
};



//-------//-------//-------//-------//-------//-------//-------//-------
//
// render
//
//-------//-------//-------//-------//-------//-------//-------//-------
static float g_time    = 0;
static int   g_wall    = 0;

struct param {
  float dist;
  vec3 n;
};

void map(param *p, vec3 pos, vec3 dir) {
  vec3 start = pos;
  vec3 t = pos;
  float d = 1.0f;
  float n = 0.0f;
  float Z = 3.0f;
  param prm;
  for(int i = 50; i-- && d > 0.003f; ) {
    d /= Z;
    t = t.frac() * Z;
    int j = int(t.integer().dot(t.integer())) % 4;
    if(j >= 2) {
      t    = (((dir > 0.0f) - t.frac()) / dir);
      if(g_wall) {
        n    = t.x;
      } else {
        n    = min(min(t.x, t.y), t.z);
      }
      pos += (dir * (n * d + 0.001f));
      //n    = dir.dot(t > n);
      t    = pos;
      d    = 1;
    }
  }
  p->dist = (pos - start).length();
  p->n = n;
}

void Update(unsigned long *dest, int width, int height) {
  g_time += 0.016666f;
  
  float a      = g_time;
  float n      = a * 0.1f;
  float v      = 0.14f;
  vec3 pos     = vec3(0.5f + sin(a * 0.2f) * v, 0.5f + cos(a * 0.3f) * v, 2.0f * a * v + sin(a * v - 0.1f)).left();
  float sn     = sin(n);
  float cn     = cos(n);
  float aspect = float(width) / float(height);
  float dx     = aspect * 2.0f / float(width);
  float dy     =          2.0f / float(height);
  float sy     = -1.0f;
  g_wall       = (int(g_time) % 30) >= 22;
  
  
  BOOL bEffect = TRUE;
  if(GetAsyncKeyState(VK_SPACE) & 0x8000) bEffect = FALSE;
  {
    for(int j = 0 ; j < height; j++) {
      float sx = -1.0f;
#ifdef _OPENMP
      #pragma omp parallel for schedule(dynamic, 1)
#endif
      for(int i = 0 ; i < width; i++) {
#ifdef _OPENMP
        sx = -1.0 + (float(i) * aspect * 2.0f / float(width));
#endif
        vec3 dir = vec3(sx, sy, 1.0f).normalize();
        dir.right() = vec3(cn * dir.x + sn * dir.z, dir.y, cn * dir.z - sn * dir.x);
        dir.right();
        dir.right() = vec3(cn * dir.x + sn * dir.z, dir.y, cn * dir.z - sn * dir.x);
        param p;
        map(&p, pos, dir);
        unsigned long result = ((dir * 0.1f) + vec3(0.4f * p.n.x * p.dist + p.dist * 0.3f)).color();
        if(!bEffect) {
          dest[i + j * width] = result;
        } else {
          color a(dest[i + j * width]);
          color b(result);
          //a.blend(b, 200);
          a.blend(b, 256, 230, 210, 180);
          dest[i + j * width] = a.argb;
        }
#ifndef _OPENMP
        sx += dx;
#endif
      }
      sy += dy;
    }
  }
  
  //anti
  if(bEffect) {
    for(int j = 1 ; j < height-1; j++) {
      for(int i = 1 ; i < width-1; i++) {
        color a(dest[(i + 0) + (j + 0) * width]);
        color b(dest[(i + 1) + (j + 0) * width]);
        color c(dest[(i - 1) + (j + 0) * width]);
        color d(dest[(i + 0) + (j + 1) * width]);
        color e(dest[(i + 0) + (j - 1) * width]);
        a.blend(b, 128);
        a.blend(c, 128);
        a.blend(d, 128);
        a.blend(e, 128);
        dest[i + j * width] = a.argb;
      }
    }
  }
  
  //fps : http://www.t-pot.com/program/13_fps/index.html
  static DWORD    last = timeGetTime();
  static DWORD    frames = 0;
  static char     buf[256] = "";
  DWORD           current;
  current = timeGetTime();
  frames++;
  if(1000 <= current - last) {
    float dt = (float)(current - last) / 1000.0f;
    float fps = (float)frames / dt;
    last = current;
    frames = 0;
    sprintf(buf, "%.02f fps", fps);
    printf("%s\n", buf);
  }
}


//-------//-------//-------//-------//-------//-------//-------//-------
// entry point
//-------//-------//-------//-------//-------//-------//-------//-------
int main(int argc, char *argb) {
  start("Create DIB", 240, 180, 0.8f, Update);
}


